@@ -15,11 +15,46 @@ class Algorithms:
     def __init__(self, verbose=False):
         """
         Initialize self. If verbose is True, progress information will be printed
-        to stdout.
+        to stdout, unless the per-function verbose flag is set to False.
         """
         self.verbose = verbose
+        self._verbose = verbose
 
-    def downsample(self, frame, iterations=1):
+    def clip(self, frame, lo=0.0, hi=1.0, verbose=True):
+        """
+        Clip all pixels in the given frame to [lo, hi]. The frame may be in either
+        RGB or raw format.
+        """
+        t0 = time.time()
+        frame_out = np.clip(frame, lo, hi)
+        self._verbose = self.verbose and verbose
+        self._vprint(f"{_elapsed(t0)} - clipping from {_minmax(frame)} to [{lo:.2f}, {hi:.2f}]")
+        return frame_out
+
+    def bayer_split(self, frame):
+        """
+        Split the given Bayer frame into four single-color frames.
+        """
+        # pylint: disable=no-self-use
+        ch1 = frame[0::2, 0::2]
+        ch2 = frame[0::2, 1::2]
+        ch3 = frame[1::2, 0::2]
+        ch4 = frame[1::2, 1::2]
+        return ch1, ch2, ch3, ch4
+
+    def bayer_combine(self, ch1, ch2, ch3, ch4):
+        """
+        Interleave the given Bayer channels into a complete frame.
+        """
+        # pylint: disable=no-self-use
+        frame = np.zeros(np.array(ch1.shape) * 2, dtype=ch1.dtype)
+        frame[0::2, 0::2] = ch1
+        frame[0::2, 1::2] = ch2
+        frame[1::2, 0::2] = ch3
+        frame[1::2, 1::2] = ch4
+        return frame
+
+    def downsample(self, frame, iterations=1, verbose=True):
         """
         Downsample the given RGB/Bayer frame by a factor of two in both directions,
         that is, to a quarter of its original size. This is repeated a given number
@@ -35,22 +70,19 @@ class Algorithms:
             if frame.ndim == 2:  # Bayer mode
                 mode = "Bayer"
                 for _ in range(iterations):
-                    imgh, imgw = frame.shape[:2]
-                    ch1 = frame[0::2, 0::2][::2, ::2]
-                    ch2 = frame[0::2, 1::2][::2, ::2]
-                    ch3 = frame[1::2, 0::2][::2, ::2]
-                    ch4 = frame[1::2, 1::2][::2, ::2]
-                    frame = np.zeros(shape=(imgh // 2, imgw // 2), dtype=frame.dtype)
-                    frame[0::2, 0::2] = ch1
-                    frame[0::2, 1::2] = ch2
-                    frame[1::2, 0::2] = ch3
-                    frame[1::2, 1::2] = ch4
+                    ch1, ch2, ch3, ch4 = self.bayer_split(frame)
+                    ch1 = ch1[::2, ::2]
+                    ch2 = ch2[::2, ::2]
+                    ch3 = ch3[::2, ::2]
+                    ch4 = ch4[::2, ::2]
+                    frame = self.bayer_combine(ch1, ch2, ch3, ch4)
             imgh, imgw = frame.shape[:2]
             factor = f"{2**iterations} x {2**iterations}"
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - {mode} downsampling [{factor}] from {orgw} x {orgh} to {imgw} x {imgh}")
         return frame
 
-    def resize(self, frame, target_width, target_height):
+    def resize(self, frame, target_width, target_height, verbose=True):
         """
         Resize the given RGB frame to the given target width and height using Lanczos
         interpolation. If target width and height are the same as current width and
@@ -62,10 +94,11 @@ class Algorithms:
         if (dstw, dsth) != (orgw, orgh):
             frame = cv2.resize(frame, (dstw, dsth), cv2.INTER_LANCZOS4)
             dsth, dstw = frame.shape[:2]
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - resizing [Lanczos] from {orgw} x {orgh} to {dstw} x {dsth}")
         return frame
 
-    def linearize(self, frame, blacklevel=None, whitelevel=None, num_clipped=1000):
+    def linearize(self, frame, blacklevel=None, whitelevel=None, num_clipped=1000, verbose=True):
         """
         Linearize the given frame such that pixels are first clipped to the range
         [BL, WL] and then remapped to [0, 1], where BL and WL are the given black
@@ -79,27 +112,32 @@ class Algorithms:
             t0 = time.time()
             percentile = num_clipped / frame.size * 100.0
             blacklevel = np.percentile(frame, percentile)
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - estimating black level: {percentile:5.2f}th percentile = {blacklevel:.2f}")
         if whitelevel is None:
             t0 = time.time()
             percentile = (1.0 - num_clipped / frame.size) * 100.0
             whitelevel = np.percentile(frame, percentile)
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - estimating white level: {percentile:5.2f}th percentile = {whitelevel:.2f}")
         if (blacklevel, whitelevel) != (0, 1):
+            frame = self.clip(frame, blacklevel, whitelevel, verbose)
             t0 = time.time()
-            frame = np.clip(frame, blacklevel, whitelevel)
             frame -= blacklevel
             frame /= whitelevel - blacklevel
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - linearizing from [{blacklevel:.2f}, {whitelevel:.2f}] to [0, 1]")
         return frame
 
-    def demosaic(self, frame, bayer_pattern):
+    def demosaic(self, frame, bayer_pattern, verbose=True):
         """
         Demosaic the given sensor raw frame using the Edge Aware Demosaicing (EAD)
         algorithm. Bayer order must be specified by the caller, and must be "RGGB",
-        "GBRG", "BGGR", or "GRBG".
+        "GBRG", "BGGR", or "GRBG". The frame must be in floating-point format with
+        all pixel values in the [0, 1] range.
         """
         t0 = time.time()
+        assert np.all((frame >= 0.0) * (frame <= 1.0)), "demosaic() requires pixel values in range [0, 1]"
         bayer_to_cv2 = {"RGGB": cv2.COLOR_BAYER_BG2RGB_EA,
                         "GBRG": cv2.COLOR_BAYER_GR2RGB_EA,
                         "BGGR": cv2.COLOR_BAYER_RG2RGB_EA,
@@ -107,10 +145,11 @@ class Algorithms:
         frame = (frame * 65535).astype(np.uint16)
         frame = cv2.cvtColor(frame, bayer_to_cv2[bayer_pattern.upper()])
         frame = frame / 65535.0
+        self._verbose = self.verbose and verbose
         self._vprint(f"{_elapsed(t0)} - demosaicing [EAD, {bayer_pattern}]: range = {_minmax(frame)}")
         return frame
 
-    def lsc(self, frame, lscmap):
+    def lsc(self, frame, lscmap, verbose=True):
         """
         Multiply the given RGB/raw frame by the given lens shading correction (LSC)
         map. If the frame is in Bayer raw format, the LSC map must have the same
@@ -140,15 +179,12 @@ class Algorithms:
                 if lscmap.ndim == 3:  # RGB
                     gains = np.amax(lscmap, axis=(0, 1))
                 if lscmap.ndim == 2:  # assume Bayer raw, ignore grayscale
-                    c1 = lscmap[::2, ::2]
-                    c2 = lscmap[::2, 1::2]
-                    c3 = lscmap[1::2, ::2]
-                    c4 = lscmap[1::2, 1::2]
-                    gains = np.array([np.amax(c) for c in (c1, c2, c3, c4)])
+                    gains = np.array([np.amax(c) for c in self.bayer_split(lscmap)])
+                self._verbose = self.verbose and verbose
                 self._vprint(f"{_elapsed(t0)} - applying LSC with max gains {gains}: range = {_minmax(frame)}")
         return frame
 
-    def wb(self, frame, gains):
+    def wb(self, frame, gains, verbose=True):
         """
         Multiply the RGB channels of the given frame by the given white balance
         coefficients. If there are only two coefficients instead of three, they
@@ -161,10 +197,11 @@ class Algorithms:
             gains = np.insert(gains, 1, 1.0) if len(gains) == 2 else gains
             frame = frame * gains
             with np.printoptions(formatter={'float': lambda x: f"{x:.3f}"}):
+                self._verbose = self.verbose and verbose
                 self._vprint(f"{_elapsed(t0)} - applying WB gains {gains}: range = {_minmax(frame)}")
         return frame
 
-    def ccm(self, frame, matrix):
+    def ccm(self, frame, matrix, verbose=True):
         """
         Apply the given color correction matrix on the given RGB frame. Input
         colors are clipped to [0, 1] to avoid "pink sky" artifacts caused by the
@@ -173,15 +210,16 @@ class Algorithms:
         recovery. If matrix is None, the frame is returned untouched.
         """
         if matrix is not None:
+            frame = self.clip(frame, 0, 1, verbose=False)
             t0 = time.time()
-            frame = np.clip(frame, 0, 1)
             frame = np.matmul(frame, matrix.T)
             with np.printoptions(formatter={'float': lambda x: f"{x:.2f}"}):
                 sums = f"with column sums {np.sum(matrix, axis=0).T}"
+                self._verbose = self.verbose and verbose
                 self._vprint(f"{_elapsed(t0)} - applying CCM {sums}: range = {_minmax(frame)}")
         return frame
 
-    def tonemap(self, frame, mode="Reinhard"):
+    def tonemap(self, frame, mode="Reinhard", verbose=True):
         """
         Apply Reinhard tonemapping on the given RGB frame, compressing the range
         [0, N] to [0, 1]. Negative values are clipped to zero. This algorithm is
@@ -189,15 +227,16 @@ class Algorithms:
         """
         if mode is not None:
             if mode == "Reinhard":
+                frame = self.clip(frame, 0, np.inf, verbose)
                 t0 = time.time()
-                frame = np.maximum(frame, 0)
                 frame = frame.astype(np.float32)  # can't handle float64
                 algo = cv2.createTonemapReinhard(gamma=1.0, intensity=0.0, light_adapt=0.0, color_adapt=0.0)
                 frame = algo.process(frame)
+                self._verbose = self.verbose and verbose
                 self._vprint(f"{_elapsed(t0)} - tonemapping [{mode}]: range = {_minmax(frame)}")
         return frame
 
-    def chroma_denoise(self, frame, strength=6, winsize=17):
+    def chroma_denoise(self, frame, strength=6, winsize=17, verbose=True):
         """
         Apply non-local means denoising (Buades et al. 2011) on the given RGB frame.
         Input colors are clipped to [0, 1] prior to denoising. Increasing the values
@@ -205,16 +244,17 @@ class Algorithms:
         and more time-consuming. If strength is 0, the frame is returned untouched.
         """
         if strength > 0:
-            t0 = time.time()
             maxval, dtype = (255, np.uint8)  # OpenCV denoising can't handle 16-bit color
-            frame = np.clip(frame * maxval + 0.5, 0, maxval)
+            frame = self.clip(frame * maxval + 0.5, 0, maxval, verbose)
+            t0 = time.time()
             frame = frame.astype(dtype)
             frame = cv2.fastNlMeansDenoisingColored(frame, h=0, hColor=strength, searchWindowSize=winsize)
             frame = frame.astype(np.float32) / maxval
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - chroma denoising [s={strength:.2f}, w={winsize}]: range = {_minmax(frame)}")
         return frame
 
-    def gamma(self, frame, mode="sRGB", lut=None):
+    def gamma(self, frame, mode="sRGB", lut=None, verbose=True):
         """
         Apply rec709 or sRGB gamma or a custom tone curve on the given frame.
         In "LUT" mode, a look-up table with N entries must be provided, where
@@ -222,9 +262,9 @@ class Algorithms:
         format-agnostic. If mode is None, the frame is returned untouched.
         """
         if mode is not None:
-            t0 = time.time()
             realmode = mode
-            frame = np.clip(frame, 0, 1)  # can't handle values outside of [0, 1]
+            frame = self.clip(frame, 0, 1, verbose)  # can't handle values outside of [0, 1]
+            t0 = time.time()
             if realmode in ["sRGB", "rec709"]:
                 bpp = 14
                 N = 2 ** bpp
@@ -245,28 +285,28 @@ class Algorithms:
                 lut = lut * maxval
             if mode == "LUT":
                 maxval = len(lut) - 1
-                verbose, self.verbose = self.verbose, False
-                frame = self.quantize(frame, maxval)  # [0, 1] ==> [0, maxval]
-                frame = lut[frame]                    # [0, maxval] ==> [0, maxval]
-                frame = frame / maxval                # [0, maxval] ==> [0, 1]
-                self.verbose = verbose
+                frame = self.quantize(frame, maxval, verbose=False)  # [0, 1] ==> [0, maxval]
+                frame = lut[frame]                                   # [0, maxval] ==> [0, maxval]
+                frame = frame / maxval                               # [0, maxval] ==> [0, 1]
+            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - applying gamma curve [{realmode}]: range = {_minmax(frame)}")
         return frame
 
-    def quantize(self, frame, maxval, dtype=np.uint16):
+    def quantize(self, frame, maxval, dtype=np.uint16, verbose=True):
         """
         Clip the given frame to [0, 1], rescale it to [0, maxval], and convert
         it to the given dtype with proper rounding. This algorithm is format-
         agnostic.
         """
         t0 = time.time()
-        frame = np.clip(frame * maxval + 0.5, 0, maxval)
+        frame = self.clip(frame * maxval + 0.5, 0, maxval, verbose=False)
         frame = frame.astype(dtype)
-        self._vprint(f"{_elapsed(t0)} - clipping to [0, 1] and rescaling to [0, {maxval}]")
+        self._verbose = self.verbose and verbose
+        self._vprint(f"{_elapsed(t0)} - quantizing from float [0, 1] to {np.dtype(dtype).name} [0, {maxval}]")
         return frame
 
     def _vprint(self, message, **kwargs):
-        if self.verbose:
+        if self._verbose:
             print(message, **kwargs)
 
 
